/* 
Name: BTFtools.cpp
Version: 1.0

Description: Additional functions for BTFBASE loading into texture rep.
Author: Jiri Filip, UTIA AS CR, filipj'at'utia'dot'cas'dot'cz
Date: 28/05/07

COPYRIGHT ======================================================================

This source code is release under Creative Commons licence
(Attribution-NonCommercial 3.0 Unported) specified here:

http://creativecommons.org/licenses/by-nc/3.0/
http://creativecommons.org/licenses/by-nc/3.0/legalcode

You are free:
    * to Share - to copy, distribute and transmit the work
    * to Remix - to adapt the work
Under the following conditions:
    * Attribution - You must attribute the work in the manner specified by the
      author or licensor (but not in any way that suggests that they endorse you
      or your use of the work).

      Attribute this work:
      ------------------------------------------------
      Bidirectional Texture Function Compression based on Multi-Level Vector 
      Quantization
      as the paper written by
          Vlastimil Havran 	Jiri Filip 	Karol Myszkowski

      and published in Computer Graphics Forum, Volume 29, Issue 1, 
      pages 175-190, March 2010, DOI: 10.1111/j.1467-8659.2009.01585.x

      Project page:
        http://www.cgg.cvut.cz/members/havran/btfbase/
      ------------------------------------------------
     
    * Noncommercial - You may not use this work for commercial purposes.

With the understanding that:
    * Waiver - Any of the above conditions can be waived if you get permission 
      from the copyright holder.
    * Public Domain - Where the work or any of its elements is in the public 
      domain under applicable law, that status is in no way affected by 
      the license.
    * Other Rights - In no way are any of the following rights affected by the 
      license:
          o Your fair dealing or fair use rights, or other applicable copyright 
            exceptions and limitations;
          o The author's moral rights;
          o Rights other persons may have either in the work itself or in how 
            the work is used, such as publicity or privacy rights.
    * Notice - For any reuse or distribution, you must make clear to others 
         the license terms of this work. The best way to do this is with a link 
         to this web page.
=========================================================================== */

#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <string>
#include <cassert>

#include "TAlloc.h"
#include "BTFtools.h"
#include "readpng.h"

#define FLOATMAX ((float)(1e30))
#define INTMAX 99999999
#define PI 3.14159265358979323846

int invertMatrix(float* mat, float* im)
{
	float rmat[4][8];
	float a, b;
	int i, j, k;

	// initialize reduction matrix
	rmat[0][0] = mat[0];
	rmat[1][0] = mat[1];
	rmat[2][0] = mat[2];
	rmat[3][0] = mat[3];
	rmat[0][1] = mat[4];
	rmat[1][1] = mat[5];
	rmat[2][1] = mat[6];
	rmat[3][1] = mat[7];
	rmat[0][2] = mat[8];
	rmat[1][2] = mat[9];
	rmat[2][2] = mat[10];
	rmat[3][2] = mat[11];
	rmat[0][3] = mat[12];
	rmat[1][3] = mat[13];
	rmat[2][3] = mat[14];
	rmat[3][3] = mat[15];
	rmat[0][4] = 1.0f;
	rmat[1][4] = 0.0f;
	rmat[2][4] = 0.0f;
	rmat[3][4] = 0.0f;
	rmat[0][5] = 0.0f;
	rmat[1][5] = 1.0f;
	rmat[2][5] = 0.0f;
	rmat[3][5] = 0.0f;
	rmat[0][6] = 0.0f;
	rmat[1][6] = 0.0f;
	rmat[2][6] = 1.0f;
	rmat[3][6] = 0.0f;
	rmat[0][7] = 0.0f;
	rmat[1][7] = 0.0f;
	rmat[2][7] = 0.0f;
	rmat[3][7] = 1.0f;

	// perform reductions
	for(i=0; i<4; i++){
		a = rmat[i][i];
		if(a == 0.0f)  // matrix is singular, can't be reduced
			return 0;
		else{
			a = 1.0f / a;
			for(j=0; j<8; j++)
				rmat[i][j] = rmat[i][j] * a;
			for(k=0; k<4; k++){
				if((k-i) != 0){
					b = rmat[k][i];
					for(j=0; j<8; j++)
						rmat[k][j] = rmat[k][j] - b * rmat[i][j];
				}
			}
		}
	}

	// extract the inverted matrix
	im[0] = rmat[0][4];
	im[1] = rmat[1][4];
	im[2] = rmat[2][4];
	im[3] = rmat[3][4];
	im[4] = rmat[0][5];
	im[5] = rmat[1][5];
	im[6] = rmat[2][5];
	im[7] = rmat[3][5];
	im[8] = rmat[0][6];
	im[9] = rmat[1][6];
	im[10] = rmat[2][6];
	im[11] = rmat[3][6];
	im[12] = rmat[0][7];
	im[13] = rmat[1][7];
	im[14] = rmat[2][7];
	im[15] = rmat[3][7];

	return 1;
}//--- invertMatrix -------------------------------------------------

void vec_diff(float *x, float *y, float *res)
{
    res[0] = x[0] - y[0];
    res[1] = x[1] - y[1];
    res[2] = x[2] - y[2];
}//--- vec_diff ----------------------------------------------------

float vec_len(float *x)
{
    return sqrtf(x[0]*x[0] + x[1]*x[1] + x[2]*x[2]);
}//--- vec_len ----------------------------------------------------

int vec_norm(float *x)
{
    float aux;
    aux = vec_len(x);
    if(aux == 0.f)
      {		
        x[0] = 0.f;
        x[1] = 0.f;
        x[2] = 0.f;
        printf("vec_norm: zero length vector");    
        return 0;
      }
    x[0] = x[0]/aux;
    x[1] = x[1]/aux;
    x[2] = x[2]/aux;
    return 1;
}//--- vec_norm ---------------------------------------------------

void vec_crossprod(float *x, float *y, float *res)
{
    res[0] = x[1]*y[2] - x[2]*y[1];
    res[1] = x[2]*y[0] - x[0]*y[2];
    res[2] = x[0]*y[1] - x[1]*y[0];
}//--- vec_crossprod ----------------------------------------------



//#################################################################################################
//### class BTFtools ########################################################JF 05/2007############
//#################################################################################################

BTFtools::BTFtools(int texSize, int SlicesPerPhi, int SlicesPerTheta, int SlicesPerHemi, int LengthOfSlice)
{
  this->texSize = texSize;
  matName = callocation1(0,1000);
  iPath = callocation1(0,1000);

  this->SlicesPerPhi   = SlicesPerPhi;
  this->SlicesPerTheta = SlicesPerTheta;
  this->SlicesPerHemi  = SlicesPerHemi;
  this->LengthOfSlice  = LengthOfSlice;

  stepBeta  = (180.f/(float)(LengthOfSlice-1)) *(PI/180.f); 
  stepAlpha = (180.f/(float)(SlicesPerHemi-1)) *(PI/180.f);
  stepPhi   = (360.f/(float)(SlicesPerPhi)) *(PI/180.f);
  stepTheta = (90.f/(float)(SlicesPerTheta-1)) *(PI/180.f);

  TI16_1 = sallocation1(0,4*texSize*texSize-1);
  TI16_2 = sallocation1(0,4*texSize*texSize-1);
  TI16_3 = sallocation1(0,4*texSize*texSize-1);
  TI16_4 = sallocation1(0,4*texSize*texSize-1);

  TF32_scales = allocation1(0,4*texSize*texSize-1);
  TF32_data   = allocation1(0,4*texSize*texSize-1);

}//--- BTFtools ---------------------------------------------------

BTFtools::~BTFtools()
{
  freecmem1(matName, 0,1000);
  freecmem1(iPath, 0,1000);

  deallocTextures();

}//--- ~BTFtools --------------------------------------------------

void BTFtools::deallocTextures()
{
  if(TI16_1!=0)
    freesmem1(TI16_1, 0,4*texSize*texSize-1);
  if(TI16_2!=0)
    freesmem1(TI16_2, 0,4*texSize*texSize-1);
  if(TI16_3!=0)
    freesmem1(TI16_3, 0,4*texSize*texSize-1);
  if(TI16_4!=0)
    freesmem1(TI16_4, 0,4*texSize*texSize-1);
  if(TF32_scales!=0)
    freemem1(TF32_scales, 0,4*texSize*texSize-1);
  if(TF32_data!=0)
    freemem1(TF32_data,   0,4*texSize*texSize-1);
}//--- deallocTextures --------------------------------------------

void
BTFtools::setMaterial(char *iPath, float &BTFwidth, float &BTFheight, float &hdrMultiplier)
// load material BTF sample parameters into textures
{
  /*
   |        indices - GL_RGBA16        |   GL_RGBA16(32)F_ARB   |
   | TI16_1 | TI16_2 | TI16_3 | TI16_4 | TF32_scales| TF32_data |
   |--------|--------|--------|--------|------------|-----------| 
  R| P6.x   | P3.x   | M.x    | I2.x   |   P6       |   P1      |
  G| P6.y   | P3.y   | M.y    | I2.y   |   P4       |   C       |
  B| P4.x   | P2.x   | empty  | I1.x   |   P3       |   empty   |
  A| P4.y   | P2.y   | empty  | I1.y   |   P2       |   empty   | 
*/

  this->iPath = iPath;
  this->matName = matName;

  char fileName[1000];
  sprintf(fileName,"%s/all_materialInfo.txt",iPath);

  FILE *fpr;
  if ((fpr=fopen(fileName,"r"))==0) 
    {
      printf("Error opening file '%s'!!!\nExiting...\n",fileName);
      exit(0);
    }
  // parsing material info file
  char matName[100];
  char auxString[1000];
  int auxInt;
  fscanf(fpr,"%s\n",auxString);
  fscanf(fpr,"%d\n",&auxInt);
  fscanf(fpr,"%d\n",&auxInt);
  fscanf(fpr,"%d\n",&auxInt);
  fscanf(fpr,"%d\n",&auxInt);
  fscanf(fpr,"%d\n",&auxInt);
  fscanf(fpr,"%d\n",&auxInt);
  fscanf(fpr,"%d\n",&auxInt);
  fscanf(fpr,"%d\n",&auxInt);
  fscanf(fpr,"%d\n",&auxInt);
  fscanf(fpr,"%d %d %d %d %d %d %d %d %d %d %d %d\n",&auxInt,&LengthOfSlice,&auxInt,&SlicesPerHemi,&auxInt,&SlicesPerTheta,&auxInt,&SlicesPerPhi, &auxInt,&auxInt,&auxInt,&auxInt);
  printf("Angular quantization: beta: %d, alpha: %d, theta: %d, phi:%d\n",LengthOfSlice,SlicesPerHemi,SlicesPerTheta,SlicesPerPhi); // 13,13,6,16
  for(int i=0;i<10;i++)
    fscanf(fpr,"%d ",&auxInt);
  fscanf(fpr,"\n");
  int sP1,sP2,sM,sP3,sP4,sI2,sI1,sC;
  fscanf(fpr,"%d %d %d %d %d %d %d %d %d %d\n",&sP1,&sC,&sI1,&sM,&sP2,&sI2,&sP3,&auxInt,&sP4,&auxInt);
  printf("Code-books sizes:   P1: %d, C: %d, I1: %d, M: %d, P2: %d, I2: %d, P3: %d, P4: %d\n",sP1,sC,sI1,sM,sP2,sI2,sP3,sP4); 
  printf("HW depend. maximas: P1: %d, C: %d, I1: %d, M: %d, P2: %d, I2: %d, P3: %d, P4: %d\n",texSize*(texSize/LengthOfSlice),texSize*(texSize/2),texSize*(texSize/LengthOfSlice),texSize*(texSize/2),texSize*(texSize/SlicesPerHemi),texSize*(texSize/SlicesPerHemi),texSize*(texSize/SlicesPerTheta),texSize*(texSize/SlicesPerPhi)); 
  float auxFloat;
  fscanf(fpr,"%d ",&auxInt);
  for(int i=0;i<11;i++)
    fscanf(fpr,"%f ",&auxFloat);
  fscanf(fpr,"\n");
  int hP6,wP6;
  float hdrMult;
  fscanf(fpr,"%s %s %s %s %d %d %d %d %f\n",&matName,&auxString,&auxString,&auxString,&auxInt,&auxInt,&hP6,&wP6,&hdrMult);
  printf("BTF size, i.e. size of code-book P6 (HW max.): %d x %d (%d x %d)\n",hP6,wP6, texSize,texSize); 
  fclose(fpr);
  printf("Material name: >>> %s <<<\n",matName);
  if (hdrMult != 0.0) {
    printf("Material is HDR with multiplier = %f\n", hdrMult);
    hdrMultiplier = 1.0f/hdrMult;
  }
  else {
    printf("Material is LDR.\n");
	hdrMultiplier = 0.0f;
  }

  BTFwidth  = wP6;
  BTFheight = hP6;

  printf("Loading code-books from '%s':\n",iPath);
  int nr,nc,maxVal,minVal,minValFloat,maxValFloat;
  int valInt, k;
  float valFloat;
  
  // PDF6D loading ================================================
  char CBname[100];
  sprintf(CBname,"%s_PDF6Dslices.txt",matName);
  sprintf(fileName,"%s/%s",iPath,CBname);
  if ((fpr=fopen(fileName,"r"))==0) 
    {
      printf("Error opening file '%s'!!!\nExiting...\n",fileName);
      exit(0);
    }
  fscanf(fpr,"%d %d %d %d\n",&nr,&nc,&minVal,&maxVal);
  for(int i=0;i<nr;i++)
    {
      for(int j=0;j<nc;j++)
        {
          if(i>=texSize || j>=texSize)
            {
              printf("Texture size limit (%dx%d) exceeded !!!\nExiting...\n",texSize,texSize);
              printf("(lower code-book size or increase texture-size, if supported by your HW)\n");
              exit(0);
            }
          fscanf(fpr,"%d ",&valInt);
          k = valInt/texSize;
          TI16_1[4*(i*texSize+j)  ] = (unsigned short)(k*SlicesPerPhi);
          TI16_1[4*(i*texSize+j)+1] = (unsigned short)(valInt%texSize);
		  //printf("%d - %d %d - %d %d %d\n",valInt,TI16_1[4*(i*texSize+j)],TI16_1[4*(i*texSize+j)+1],k,SlicesPerPhi,texSize);
        }
      fscanf(fpr,"\n");
    }
  fclose(fpr);
  printf("  - indices '%s' loaded\n",CBname);

  sprintf(CBname,"%s_PDF6Dscale.txt",matName);
  sprintf(fileName,"%s/%s_PDF6Dscale.txt",iPath,matName);
  if ((fpr=fopen(fileName,"r"))==0) 
    {
      printf("Error opening file '%s'!!!\nExiting...\n",fileName);
      exit(0);
    }
  fscanf(fpr,"%d %d %f %f\n",&nr,&nc,&minValFloat,&maxValFloat);
  for(int i=0;i<nr;i++)
    {
      for(int j=0;j<nc;j++)
        {
          fscanf(fpr,"%f ",&valFloat);
          TF32_scales[4*(i*texSize+j)] = valFloat;
        }
      fscanf(fpr,"\n");
    }
  fclose(fpr);
  printf("  - scales  '%s' loaded\n",CBname);

  // luminance channel only ##############
  // P4 code-book loading ================================================
  loadCodeBookIndexScale("all_PDF4Dslices.txt", "all_PDF4Dscale.txt", SlicesPerPhi, SlicesPerTheta, 
                         TI16_1, 2, TF32_scales, 1);

  // P3 code-book loading ================================================
  loadCodeBookIndexScale("all_PDF3Dslices.txt", "all_PDF3Dscale.txt", SlicesPerTheta, 2,
                         TI16_2, 0, TF32_scales, 2);

  // M code-book loading =================================================
  loadCodeBookIndex("all_PDF2Dindices.txt", 2, SlicesPerHemi, TI16_3, 0);

  // P2 code-book loading ================================================
  loadCodeBookIndexScale("all_PDF2Dslices.txt", "all_PDF2Dscale.txt", SlicesPerHemi, LengthOfSlice,
                         TI16_2, 2, TF32_scales, 3);

  // P1 code-book loading ================================================
  loadCodeBookScale("all_PDF1Dslice.txt", LengthOfSlice, TF32_data, 0);

  // colour components a-b ###############
  // I2 code-book loading ================================================
  loadCodeBookIndex("all_PDF2Dcolours.txt", SlicesPerHemi, LengthOfSlice, TI16_4, 0);

  // I1 code-book loading ================================================
  loadCodeBookIndex("all_indexAB.txt", LengthOfSlice, 2, TI16_4, 2);

  // C code-book loading =================================================
  loadCodeBookScale("all_colors.txt", 2, TF32_data, 1);
  return;
}//--- setMaterial -----------------------------------------------------------------

void BTFtools::loadCodeBookIndex(char *indicesName, int sizeX, int valSizeX, unsigned short *indices, int indexStartPlane)
// loads code-book file with indices
{
  FILE *fpr;
  int nr,nc, minVal,maxVal, Y,X,valX,valY;
  char fileName[1000];
  int i,j, valInt;

  // indices loading ==============================================
  sprintf(fileName,"%s/%s",iPath,indicesName);
  if ((fpr=fopen(fileName,"r"))==0) 
    {
      printf("Error opening file '%s'!!!\nExiting...\n",fileName);
      exit(0);
    }
  fscanf(fpr,"%d %d %d %d\n",&nr,&nc,&minVal,&maxVal);
  for(i=0;i<nr;i++)
    {
      for(j=0;j<nc;j++)
        {
          fscanf(fpr,"%d ",&valInt);
          putIndexAtUV(i,j, valInt, sizeX, valSizeX, Y,X, valY,valX);
          if(Y>=texSize || X>=texSize)
            {
               printf("Texture size limit (%dx%d) exceeded for code-book: '%s'!!!\n",texSize,texSize,indicesName);
               printf("(lower code-book size or increase texture-size, if supported by your HW)\nExiting...\n");
               exit(0);
            }
          indices[4*(Y*texSize+X)+indexStartPlane  ] = (unsigned short)(valX);
          indices[4*(Y*texSize+X)+indexStartPlane+1] = (unsigned short)(valY);
        }
      fscanf(fpr,"\n");
    }
  fclose(fpr);
  //  printf("    %d %d - %d %d - %d - %d %d (%d %d)\n",i,j, Y,X, indexStartPlane, valY,valX, nr,nc);
  printf("  - indices '%s' loaded\n",indicesName);

  return;
}//--- loadCodeBookIndex ------------------------------------------------------------

void BTFtools::loadCodeBookScale(char *scalesName, int sizeX, float *scales, int scalePlane)
// loads code-book file with scales
{
  FILE *fpr;
  int i,j, nr,nc, Y,X;
  char fileName[1000];
  float valFloat, minVal,maxVal;

  // scales loading ===============================================
  sprintf(fileName,"%s/%s",iPath,scalesName);
  if ((fpr=fopen(fileName,"r"))==0) 
    {
      printf("Error opening file '%s'!!!\nExiting...\n",fileName);
      exit(0);
    }
  fscanf(fpr,"%d %d %f %f\n",&nr,&nc,&minVal,&maxVal);
  for(i=0;i<nr;i++)
    {
      for(j=0;j<nc;j++)
        {
          fscanf(fpr,"%f ",&valFloat);
          putIndexAtUV(i,j, sizeX, Y,X);
          if(Y>=texSize || X>=texSize)
            {
                  printf("Texture size limit (%dx%d) exceeded for code-book: '%s'!!!\n",texSize,texSize,scalesName);
                  printf("(lower code-book size or increase texture-size, if supported by your HW)\nExiting...\n");
                  exit(0);
            }
          scales[4*(Y*texSize+X)+scalePlane] = valFloat;
        }
      fscanf(fpr,"\n");
    }
  fclose(fpr);
  //  printf("    %d %d - %d %d - %d - %f (%d %d)\n",i,j, Y,X, scalePlane, valFloat,nr,nc);
  printf("  - scales  '%s' loaded\n",scalesName);

  return;
}//--- loadCodeBookScale -----------------------------------------------------------

void BTFtools::loadCodeBookIndexScale(char *indicesName, char *scalesName, int sizeX, int valSizeX, unsigned short *indices, int indexStartPlane, float *scales, int scalePlane)
{
  // indices loading =================================================
  loadCodeBookIndex(indicesName, sizeX, valSizeX, indices, indexStartPlane);

  // scales loading ==================================================
  loadCodeBookScale(scalesName, sizeX, scales, scalePlane);

  return;
}//--- loadCodeBookIndexScale ------------------------------------------------------

void BTFtools::putIndexAtUV(int iY, int iX, int val, int sizeX, int valSizeX, int &Y, int &X, int &valY, int &valX)
// computes approprite UV index for parameters placement in the texture
{
  int k = iY / texSize;
  X = k * sizeX + iX;
  Y = iY % texSize;

  k = val / texSize;
  valY = val % texSize;
  valX = k * valSizeX;
}//--- putIndexAtUV ----------------------------------------------------------------

void BTFtools::putIndexAtUV(int iY, int iX, int sizeX, int &Y, int &X)
// computes approprite UV index for parameters placement in the texture
{
  int k = iY / texSize;
  X = k * sizeX + iX;
  Y = iY % texSize;
}//--- putIndexAtUV ----------------------------------------------------------------
