/* 
Name: btfbaseGPU.cpp
Version: 1.0

Description: Main file for real-time rendering of BTFBASE
Author: Jiri Filip, UTIA AS CR, filipj'at'utia'dot'cas'dot'cz
Date: 22/11/10

COPYRIGHT ======================================================================

This source code is release under Creative Commons licence
(Attribution-NonCommercial 3.0 Unported) specified here:

http://creativecommons.org/licenses/by-nc/3.0/
http://creativecommons.org/licenses/by-nc/3.0/legalcode

You are free:
    * to Share - to copy, distribute and transmit the work
    * to Remix - to adapt the work
Under the following conditions:
    * Attribution - You must attribute the work in the manner specified by the
      author or licensor (but not in any way that suggests that they endorse you
      or your use of the work).

      Attribute this work:
      ------------------------------------------------
      Bidirectional Texture Function Compression based on Multi-Level Vector 
      Quantization
      as the paper written by
          Vlastimil Havran 	Jiri Filip 	Karol Myszkowski

      and published in Computer Graphics Forum, Volume 29, Issue 1, 
      pages 175-190, March 2010, DOI: 10.1111/j.1467-8659.2009.01585.x

      Project page:
        http://www.cgg.cvut.cz/members/havran/btfbase/
      ------------------------------------------------
     
    * Noncommercial - YOU MAY NOT USE THIS WORK FOR COMMERCIAL PURPOSES!

With the understanding that:
    * Waiver - Any of the above conditions can be waived if you get permission 
      from the copyright holder.
    * Public Domain - Where the work or any of its elements is in the public 
      domain under applicable law, that status is in no way affected by 
      the license.
    * Other Rights - In no way are any of the following rights affected by the 
      license:
          o Your fair dealing or fair use rights, or other applicable copyright 
            exceptions and limitations;
          o The author's moral rights;
          o Rights other persons may have either in the work itself or in how 
            the work is used, such as publicity or privacy rights.
    * Notice - For any reuse or distribution, you must make clear to others 
         the license terms of this work. The best way to do this is with a link 
         to this web page.
 ===========================================================================*/

#pragma comment(lib,"opengl32.lib")
#pragma comment(lib,"glu32.lib")
#pragma comment(lib,"glut32.lib")
#pragma comment(lib,"glew32.lib")

#include <cstdlib>
#include "GL/glew.h"
#include "GL/glut.h"
#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include <string.h>

#include "TAlloc.h"
#include "OBJloader.h"
#include "BTFtools.h"
#include "readpng.h"

#define PI 3.14159265358979323846

int renderWire = 0;            // toggle wireframe rendering only
int drawNeedles = 0;           // toggle additional normals drawing
int whiteBcg = 1;              // switch between white / black background
int lightShiftMode = 1;        // switch between camera / light background
int animation = 0;             // allows automatic rotation of 3D object

int actualOBJ = 0;             // specifies actual 3D OBJ model
int NoOfLoadedMaterials = 0;   // initial number of loaded materials

BTFtools *BTF;                 // BTF processing class definition

//=== materials (BTFs) data loading =========================================
const int NoOfMaterials = 1;   // maximal allowed number of BTF materials

//=== OBJ data loading ======================================================
const int NoOfObjects = 2;     // maximal number of single 3D objects in virtual scene
int loadedBTFObjects = 0;      // actual number of loaded BTF covered objects
TObjectData *DATA[NoOfObjects];// class for OBJ 3D object files loading
long verticesPerObject = 50000;// maximal number of vertices per 3D object
char objectName[NoOfObjects][100];
int *objectMaterialID;         // list of materials attached to individual 3D objects

//=== texture mapping =======================================================
float *texMapSizeX;               // required horizontal size of texture mapped on 3D object
float *texMapSizeY;               // required vertical size of texture mapped on 3D object
int Npow = 10;

//=== material specific variables ===========================================
float BTFheight;                                                   
float BTFwidth;                                                    
float hdrMultiplier;

//=== cubemaps ==============================================================
enum {CUBE_POS_X, CUBE_NEG_X, CUBE_POS_Y, CUBE_NEG_Y, CUBE_POS_Z, CUBE_NEG_Z};//cube-map axes definition
GLenum cubefaces[6] = {
  GL_TEXTURE_CUBE_MAP_POSITIVE_X_EXT,
  GL_TEXTURE_CUBE_MAP_NEGATIVE_X_EXT,
  GL_TEXTURE_CUBE_MAP_POSITIVE_Y_EXT,
  GL_TEXTURE_CUBE_MAP_NEGATIVE_Y_EXT,
  GL_TEXTURE_CUBE_MAP_POSITIVE_Z_EXT,
  GL_TEXTURE_CUBE_MAP_NEGATIVE_Z_EXT,
};                             // cube-map faces definition

static GLuint cubeMaps[2];     // cubemap handles definition

//=== parametric textures ===================================================
const int texSize = 2048;
static GLuint indexTex[4]; // integer16bit index textures (1024x1024) handle 
static GLuint scaleTex[2]; // float32bit index textures (1024x1024) handle
 
//== visualisation ==========================================================
int win_w=800,win_h=600;                       // size of GLUT window
int viewport[4] = {0,0,win_h,win_w};           // definition of OpenGl viewport

float perspectiveAngle=30.;                    // definition of viewing verspective angle
int oldwinx, oldwiny;                          // variables determining mouse movement
int mb[3] = {0, 0, 0};                         // mouse buttons: left, middle, right

// initial camera position and rotation
float xyz[3];     // translation y,x,z
float hpr[3];     // rotation

float lpos[4];    // initial light position - from left
float cpos[4];  // initial camera position
float scale = 1.;                              // 3D vertex data scaling parameter

//--- FPS ---------------
long frame, Time, timebase;                    // variables for FPS computation
char FPSstring[30];                            // string for FPS info output
char modestring[30];                           // string for light mode indication
//--- screenshot number ----------------
int screenshotNumber = 1;                      // actual number of framebuffer screenshot  
char *auxString;                               // auxilary char string
//== shaders params ========================================================
GLhandleARB v_BTF, f_BTF, p_BTF;               // BTF shader program handles
GLhandleARB v_color, f_color, p_color;         // 2D interface program shader handles

//== string for version and extensions printing ============================
char *version = callocation1(0,100);
char *extension = callocation1(0,10000);

//========= the global timer ===============================================
char buf[200];
float FPS;

// === object rotation ==========================
float rotation[3]; 
float objectShift = 0.f;
float phiPerCircle = 30.f;
int OBJrotationFlag = 0;

//=== BTFBASE codebooks sizes =====================================
int SlicesPerPhi   = 16;
int SlicesPerTheta =  7;
int SlicesPerHemi  = 11;
int LengthOfSlice  = 11;

//#####################################################################################
//### FUNCTIONS #######################################################################
//#####################################################################################

void
initRenderer()
// initializes camera and light position, direction and width of view
{
  perspectiveAngle = 64.f;
  xyz[0] =    0.0;          // translation y,x,z
  xyz[1] =    0.0;
  xyz[2] = -210.0; 
  hpr[0] =    0.0;          // rotation
  hpr[1] =    0.0;
  hpr[2] =    0.0;
  lpos[0] =    0.f;          // initial light position - from left
  lpos[1] =    0.f;
  lpos[2] =  300.f;
  lpos[3] =    1.f;

  cpos[0] = -xyz[0];
  cpos[1] = -xyz[1];
  cpos[2] = -xyz[2];
  cpos[3] = 1.f;;
  return;
}//--- initRenderer --------------------------------------------------------

char* 
readProgram(const char* name)
// loads shader program -----------
{
	FILE *in = fopen(name, "r");
   
	if (!in)
          {
            printf("Error in program reading...\n");
            return 0;
          }
	char *buffer = 0;
   
	// get file size
	long size = 0, curpos;
	curpos = ftell(in);
	fseek(in, 0L, SEEK_END);
	size = ftell(in);
	fseek(in, curpos, SEEK_SET);
   
	if (!size)
          {
            printf("Error in program reading...\n");
            return 0;   
          }
	if (!(buffer = new char[size + 1]))
          {
            printf("Error in program reading...\n");
            return 0;
          }
        printf("Program length: %d bytes\n",size+1);
        // allocating program buffer
        buffer = (char*)malloc(size+1);

	memset(buffer, '\0', size + 1);
	if (!buffer)
          {
            printf("Error in program reading...\n");
            return 0;
          }
	fread(buffer, 1, size, in); // check return with size?
	fclose(in);

	return buffer;
}//--- readProgram -------------------------------------------------------

void 
printInfoLog(GLhandleARB obj, char *shaderTitle)
// prints output of the shader program compilation
{
  int infologLength = 0;
  int charsWritten  = 0;
  char *infoLog;
  
  glGetObjectParameterivARB(obj, GL_OBJECT_INFO_LOG_LENGTH_ARB, &infologLength);
  
  if (infologLength > 0)
    {      
      printf("=== %s ============================================================\n",shaderTitle);
      infoLog = (char *)malloc(infologLength);
      glGetInfoLogARB(obj, infologLength, &charsWritten, infoLog);
      printf("%s\n",infoLog);
      printf("================================================================================\n");
      free(infoLog);
    }
  return;
}//--- printInfoLog ------------------------------------------------

void 
changeSize(int w, int h) 
// procedures called on change of window size 
{
  win_h = h;
  win_w = w;
  // Prevent a divide by zero, when window is too short (you cant make a window of zero width).
  if(h == 0)
    h = 1;

  float ratio = 1.0* w / h;

  // Reset the coordinate system before modifying
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
	
  // Set the viewport to be the entire window
  glViewport(0, 0, w, h);

  // Set the correct perspective.
  gluPerspective(perspectiveAngle,ratio,0.5,5000);
  glMatrixMode(GL_MODELVIEW);
  return;
}//--- changeSize -----------------------------------------------------------

void 
rotate3DPoint(float angle, float *v, int axis)
// computes result of 3D point rotation for specified angle around selested axis
{
  float sv[3];
  // ... rotation of whole enviroment sphere by angle phiShift
  int a,b;
  if(axis == 0) {
    a = 0;
    b = 1;
  }
  if(axis == 1) {
    a = 0;
    b = 2;
  }
  if(axis == 2) {
    a = 1;
    b = 2;
  }

  sv[a] = cos(angle) * v[a] - sin(angle) * v[b];
  sv[b] = sin(angle) * v[a] + cos(angle) * v[b];
  v[a] = sv[a];
  v[b] = sv[b];
  return;
}//-- rotate3DPoint -------------------------------------------------------

void 
rotateOBJModelGeometry(int OBJmodel, int axis, float rotAngle)
// rotates 3D object geometry for specified angle around selested axis
{
  float v[3],n[3],t[3];
  for(int i=1;i<=DATA[OBJmodel]->noOfVertices();i++)
    {
      for(int j=0;j<3;j++)
        {
          v[j] = DATA[OBJmodel]->vertices[i][j];
          n[j] = DATA[OBJmodel]->normals[i][j];
          t[j] = DATA[OBJmodel]->tangentsV[i][j];
        }
      rotate3DPoint(rotAngle, v, axis);
      rotate3DPoint(rotAngle, n, axis);
      rotate3DPoint(rotAngle, t, axis);
      for(int j=0;j<3;j++)
        {
          DATA[OBJmodel]->vertices[i][j]  = v[j];
          DATA[OBJmodel]->normals[i][j]   = n[j];
          DATA[OBJmodel]->tangentsV[i][j] = t[j];
        }
    }
  return;
}//--- rotateOBJModelGeometry -----------------------------------------------

void 
translateOBJModelGeometry(int OBJmodel, int axis, float value)
// translates 3D object geometry for specified value along selected axis 
{
  for(int i=1;i<=DATA[OBJmodel]->noOfVertices();i++) // for faces
    for(int j=0;j<3;j++) // for vertices
      DATA[OBJmodel]->vertices[i][j] += value;
  return;
}//--- translateOBJModelGeometry -------------------------------------------

void
setGPUcolor(float R, float G, float B)
// resets actual GPU color
{
	glUseProgramObjectARB(p_color);
	GLint loc = glGetUniformLocationARB(p_color,"color");
	glUniform3fARB(loc,R,G,B);
	return;
}//--- setGPUcolor --------------------------------------

void
drawVector(float *X, float *V, float color[], float scale)
// draw line representing vector
{
	float Vs[3];
	for(int i=0;i<3;i++)
		Vs[i] = scale*V[i];
	glDisable(GL_TEXTURE_2D);
	setGPUcolor(color[0],color[1],color[2]);
	glBegin(GL_LINES);
		glVertex3fv(X);
		glVertex3f(X[0]+Vs[0],X[1]+Vs[1],X[2]+Vs[2]);
	glEnd();
	glEnable(GL_TEXTURE_2D);
	return;
}//--- drawVector -----------------------------------------------------------------

void 
renderObject(TObjectData *objectData)
// renders 3D geometry
{
  GLfloat v1[3],v2[3],v3[3];
  GLfloat n1[3],n2[3],n3[3];
  GLfloat t1[3],t2[3],t3[3];
  GLfloat uv1[2],uv2[2],uv3[2];
  float den;
  float T[3]; // tangent vector
  GLint loc1,loc2,loc3,loc4;
  loc1 = glGetAttribLocationARB(p_BTF,"camera");
  loc2 = glGetAttribLocationARB(p_BTF,"light");
  loc3 = glGetAttribLocationARB(p_BTF,"tangent");
  loc4 = glGetAttribLocationARB(p_BTF,"normal");

  // ... rotation of the OBJ vertices and TBN vectors by angle 'objectShift'
  float rotAngle;
  if(OBJrotationFlag>0)
    {
      if(OBJrotationFlag==1)
        rotAngle = PI/phiPerCircle;
      
      if(OBJrotationFlag==2)
        rotAngle = - PI/phiPerCircle;
      rotateOBJModelGeometry(actualOBJ, 1, rotAngle);
      if(animation==0)
        OBJrotationFlag = 0;
    }

  glBegin(GL_TRIANGLES);

  for(int i=1;i<=objectData->noOfFaces();i++)
    {
      for(int j=0;j<3;j++)
        {
		  v1[j] = scale*objectData->getV(i,0,j);
          v2[j] = scale*objectData->getV(i,1,j);
          v3[j] = scale*objectData->getV(i,2,j);
          
          n1[j] = objectData->getN(i,0,j);
          n2[j] = objectData->getN(i,1,j);
          n3[j] = objectData->getN(i,2,j);

          t1[j] = objectData->getT(i,0,j);
          t2[j] = objectData->getT(i,1,j);
          t3[j] = objectData->getT(i,2,j);
        }
      // texture UV coordinates
      for(int j=0;j<2;j++)
        {
          uv1[j] = objectData->getUV(i,0,j);
          uv2[j] = objectData->getUV(i,1,j);
          uv3[j] = objectData->getUV(i,2,j);
        }

	  // passing vertices information
      glVertexAttrib4fvARB(loc1,cpos);  
      glVertexAttrib4fvARB(loc2,lpos);  
      glVertexAttrib3fvARB(loc3,t1);
      glVertexAttrib3fvARB(loc4,n1);  
      glMultiTexCoord2fv(GL_TEXTURE0,uv1); 
      glVertex3fv(v1); 

      glVertexAttrib4fvARB(loc1,cpos);  
      glVertexAttrib4fvARB(loc2,lpos);  
      glVertexAttrib3fvARB(loc3,t2);
      glVertexAttrib3fvARB(loc4,n2);  
      glMultiTexCoord2fv(GL_TEXTURE0,uv2); 
      glVertex3fv(v2);

      glVertexAttrib4fvARB(loc1,cpos);  
      glVertexAttrib4fvARB(loc2,lpos);  
      glVertexAttrib3fvARB(loc3,t3);
      glVertexAttrib3fvARB(loc4,n3);  
      glMultiTexCoord2fv(GL_TEXTURE0,uv3); 
      glVertex3fv(v3); 

	}// for all faces

 	glEnd();

	//--- TBN axis + lightDir visualization ----------------- 
    if(drawNeedles)
	  {
		for(int i=1;i<=objectData->noOfFaces();i++)
		{
			for(int j=0;j<3;j++)
				{
					v1[j] = scale*objectData->getV(i,0,j);
					v2[j] = scale*objectData->getV(i,1,j);
					v3[j] = scale*objectData->getV(i,2,j);
          
					n1[j] = objectData->getN(i,0,j);
					n2[j] = objectData->getN(i,1,j);
					n3[j] = objectData->getN(i,2,j);

					t1[j] = objectData->getT(i,0,j);
					t2[j] = objectData->getT(i,1,j);
					t3[j] = objectData->getT(i,2,j);
				}
			// texture UV coordinates
			for(int j=0;j<2;j++)
				{
					uv1[j] = objectData->getUV(i,0,j);
					uv2[j] = objectData->getUV(i,1,j);
					uv3[j] = objectData->getUV(i,2,j);
				}

			vec_norm(n1);		
			vec_norm(t1);		
			float b1[3];
			vec_crossprod(n1, t1, b1);
			vec_norm(b1);

			float L[3],C[3];
			vec_diff(lpos,v1,L);
			vec_diff(cpos,v1,C);

			float Ldir[3],Cdir[3];
			Ldir[0] = t1[0]*L[0] + t1[1]*L[1] + t1[2]*L[2];
			Ldir[1] = b1[0]*L[0] + b1[1]*L[1] + b1[2]*L[2];
			Ldir[2] = n1[0]*L[0] + n1[1]*L[1] + n1[2]*L[2];
			Cdir[0] = t1[0]*C[0] + t1[1]*C[1] + t1[2]*C[2];
			Cdir[1] = b1[0]*C[0] + b1[1]*C[1] + b1[2]*C[2];
			Cdir[2] = n1[0]*C[0] + n1[1]*C[1] + n1[2]*C[2];
			vec_norm(Ldir);
			vec_norm(Cdir);

			float scale = 10.0;
  			float RGB[3];
			RGB[0] = 1.0; RGB[1] = 0.0;	RGB[2] = 0.0;
			drawVector(v1,n1, RGB, scale);
			RGB[0] = 0.0; RGB[1] = 1.0;	RGB[2] = 0.0;
			drawVector(v1,t1, RGB, scale);
			RGB[0] = 0.0; RGB[1] = 0.0;	RGB[2] = 1.0;
			drawVector(v1,b1, RGB, scale);
			RGB[0] = 1.0; RGB[1] = 1.0; RGB[2] = 0.0;
			Ldir[0] *= -1.f;
			Ldir[1] *= -1.f;
			Cdir[0] *= -1.f;
			Cdir[1] *= -1.f;
			drawVector(v1,Ldir, RGB, scale);
		 }
	  }
	  //--- TBN axis + lightDir visualization ----------------- 

  return;
}//--- renderObject ---------------------------------------------------------

void 
setBTFMaterial(int actualObject)
// sets parameters of compressed BTF sample and sends them to the shader
{
  //--- cluster and cluster index textures setting ---------------------------
  glActiveTexture(GL_TEXTURE3);
  GLint loc1 = glGetUniformLocationARB(p_BTF,"TI16_1");
  glUniform1iARB(loc1,3);  // Bind "TI16_1" to texture unit 3 
  glBindTexture(GL_TEXTURE_2D,indexTex[0]);

  glActiveTexture(GL_TEXTURE4);
  GLint loc2 = glGetUniformLocationARB(p_BTF,"TI16_2");
  glUniform1iARB(loc2,4);  // Bind "TI16_2" to texture unit 4 
  glBindTexture(GL_TEXTURE_2D,indexTex[1]);

  glActiveTexture(GL_TEXTURE5);
  GLint loc3 = glGetUniformLocationARB(p_BTF,"TI16_3");
  glUniform1iARB(loc3,5);  // Bind "TI16_3" to texture unit 5 
  glBindTexture(GL_TEXTURE_2D,indexTex[2]);

  glActiveTexture(GL_TEXTURE6);
  GLint loc4 = glGetUniformLocationARB(p_BTF,"TI16_4");
  glUniform1iARB(loc4,6);  // Bind "TI16_4" to texture unit 6
  glBindTexture(GL_TEXTURE_2D,indexTex[3]);

  glActiveTexture(GL_TEXTURE7);
  GLint loc5 = glGetUniformLocationARB(p_BTF,"TF32_scales");
  glUniform1iARB(loc5,7);  // Bind "TF32_scales" to texture unit 7 
  glBindTexture(GL_TEXTURE_2D,scaleTex[0]);

  glActiveTexture(GL_TEXTURE8);
  GLint loc6 = glGetUniformLocationARB(p_BTF,"TF32_data");
  glUniform1iARB(loc6,8);  // Bind "TF32_data" to texture unit 8
  glBindTexture(GL_TEXTURE_2D,scaleTex[1]);

  //--- texture mapping parameters settings ----------------------------------
  float pixStep   = 1.f/texSize;      
  float texCutX   = BTFwidth/texSize;     
  float texCutY   = BTFheight/texSize;    

  GLint loc7 = glGetUniformLocationARB(p_BTF,"texSize");
  glUniform1fARB(loc7,texSize);  
  GLint loc7a = glGetUniformLocationARB(p_BTF,"texMapSizeX");
  glUniform1fARB(loc7a,texMapSizeX[0]);  
  GLint loc7b = glGetUniformLocationARB(p_BTF,"texMapSizeY");
  glUniform1fARB(loc7b,texMapSizeY[0]);  
  GLint loc8 = glGetUniformLocationARB(p_BTF,"pixStep");
  glUniform1fARB(loc8,pixStep);  
  GLint loc9 = glGetUniformLocationARB(p_BTF,"BTFheight");
  glUniform1fARB(loc9,(float)BTFheight);  
  GLint loc10 = glGetUniformLocationARB(p_BTF,"BTFwidth");
  glUniform1fARB(loc10,(float)BTFwidth);  
  GLint loc11 = glGetUniformLocationARB(p_BTF,"texCutX");
  glUniform1fARB(loc11,texCutX);  
  GLint loc12 = glGetUniformLocationARB(p_BTF,"texCutY");
  glUniform1fARB(loc12,texCutY);  

  float SlicesPerPhi = 16.f; //??? Why 16 and not by the parameter
  GLint loc13 = glGetUniformLocationARB(p_BTF,"SlicesPerPhi");
  glUniform1fARB(loc13,SlicesPerPhi);  
  GLint loc14 = glGetUniformLocationARB(p_BTF,"hdrMultiplier");
  // HDR is not enought bright for some reason
  const float coeffBrightness = 1.6f;
  glUniform1fARB(loc14, coeffBrightness * hdrMultiplier);  
  return;
}//--- setBTFMaterial -----------------------------------------------

void 
renderScene(void)
// renders object/s and computes FPS  
{
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  //=== modelview setting ================================
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glRotatef(hpr[2], 0.0f, 0.0f, 1.0f);
  glRotatef(hpr[1], 1.0f, 0.0f, 0.0f);
  glRotatef(hpr[0], 0.0f, 1.0f, 0.0f);
  glTranslatef(xyz[0], xyz[1], xyz[2]);

  //=== scene objects rendering ======================================
  glUseProgramObjectARB(p_BTF);
  setBTFMaterial(actualOBJ);
  renderObject(DATA[actualOBJ]);

  //---- FPS computation -----------------------------------------
  frame++;
  Time=glutGet(GLUT_ELAPSED_TIME);
  if (Time - timebase > 1000) 
    {
      FPS = frame*1000.0/(Time-timebase);
      timebase = Time;		
      frame = 0;
    }
  sprintf(buf, "BTFBASE renderer:   WxH=%dx%d, %s, FPS: %4.1f - OBJ: %s", win_w, win_h, modestring, FPS, objectName[actualOBJ]);
  glutSetWindowTitle(buf);

  glutSwapBuffers();
  return;
}//--- renderScene ------------------------------------------------

void
printHelp()
{
  printf(" ________________________________________________________________________\n"
	 " |HELP on individual keys:                                              |\n"
	 " |   a|s - rotate object to left|right                                  |\n"
	 " |   d   - switch on/off animation (start by a/s)                       |\n"
	 " |   c   - print camera translation and rotation                        |\n"
	 " |   v   - use light backgrond                                          |\n"
	 " |   b   - use dark backgroud                                           |\n"
	 " |   m   - switch between camera and light mouse movement               |\n"
	 " |   o|O - loop between available OBJ 3D models (forward|backward)      |\n"
	 " |   [|] - decrease/increase texture size twice                         |\n"
	 " |   +|- - zoom in, zoom out                                            |\n"
	 " |   n   - draw normals only                                            |\n"
	 " |   e   - print OpenGL version and available extensions                |\n"
	 " |   r   - re-initialize camera and light positions                     |\n"
	 " |   g   - grab a screen-shot                                           |\n"
	 " |   w   - toggle between fill anf wire-frame display of polygons       |\n"
	 " |   h   - this help                                                    |\n"
	 " |  Esc  - quit the renderer                                            |\n"
	 " |______________________________________________________________________|\n\n");
  return;
}//--- printHelp -------------------------------------------------------

void 
processKeys(unsigned char key, int x, int y)
// callback function on keyboard events 
{
  GLuint loc;
  unsigned char *buffer = ballocation1(0,4*win_w*win_h-1);

   switch (key) {
      case 27:         // finish the renderer
        printf("\n\n Thanks for using BTFBASE GPU renderer!\n\n");
        exit(0);
        break;
      case 'c':        // print camera translation and rotation
        printf("hpr: %3.2f %3.2f %3.2f                                  \n",hpr[0],hpr[1],hpr[2]);
        printf("xyz: %3.2f %3.2f %3.2f                                  \n",xyz[0],xyz[1],xyz[2]);
        break;
      case 'r':        // reinit camera position
        initRenderer();
        glutPostRedisplay();
        break;
      case 'e':        // printout OpenGL Driver version, available extensions
        version = (char *)glGetString(GL_VERSION);
        printf("OpenGL version: %s\n",version);
        extension = (char *)glGetString(GL_EXTENSIONS);
        printf("OpenGL extensions: %s\n",extension);
        break;
      case 'g':        // grab a screenshot
        glReadPixels(0,0,win_w,win_h, GL_RGBA, GL_UNSIGNED_BYTE, buffer);
        sprintf(auxString,"../data/screenshot%02d.png",screenshotNumber);
        loc = write_PNG_file_1D(auxString, buffer,3, win_h, win_w, 1);
        printf("Writing framebuffer to '%s' file with signal %d\n",auxString,loc); 
        screenshotNumber++;
        break;
      case 'w':        // switch to wireframe model 
        if(renderWire == 1)
          {
            glPolygonMode(GL_FRONT, GL_FILL);
            renderWire = 0;
          }
        else          
          {
            glPolygonMode(GL_FRONT, GL_LINE);
            renderWire = 1;
          }
        glutPostRedisplay();
        break;
      case 'm':       // switch between camera and light movement 
        if(lightShiftMode == 1)
          {
            lightShiftMode = 0;
            strcpy(modestring,"camera mode");
            printf("Camera: x: %7.2f y: %7.2f  height: %7.2f\r",cpos[0],cpos[1],cpos[2]);fflush(stdout);
          }
        else
          {
            lightShiftMode = 1;
            strcpy(modestring,"light mode");
            printf("Light : x: %7.2f y: %7.2f  height: %7.2f\r",lpos[0],lpos[1],lpos[2]);fflush(stdout);
          }
        break;
      case 'b':       // dark background colour
        whiteBcg = 0;
        glClearColor(0.220, 0.259, 0.251, 1.0);
        glutPostRedisplay();        
        break;
      case 'v':       // light background colour
        whiteBcg = 1;
        glClearColor(0.722, 0.796, 0.780, 1.0);
        glutPostRedisplay();
        break;
      case '+':       // zoom in
        cpos[0] /= 1.5; xyz[0] = -cpos[0];
        cpos[1] /= 1.5; xyz[1] = -cpos[1];
        cpos[2] /= 1.5; xyz[2] = -cpos[2];
        glTranslatef(xyz[0], xyz[1], xyz[2]);
        break;
      case '-':       // zoom out
        cpos[0] *= 1.5; xyz[0] = -cpos[0];
        cpos[1] *= 1.5; xyz[1] = -cpos[1];
        cpos[2] *= 1.5; xyz[2] = -cpos[2];
        glTranslatef(xyz[0], xyz[1], xyz[2]);
        break;

      case '[':         // decrease BTF texture size
          if(Npow>7)
            {
              Npow--;
              texMapSizeY[0] = (int)pow(2.f, Npow);
              texMapSizeX[0] = (int)pow(2.f, Npow);
              printf("   Texture size : [%4.0f x %4.0f]                                                     \r",texMapSizeY[0],texMapSizeX[0]);fflush(stdout);
              glutPostRedisplay();
            }
        break;
      case ']':         // increase BTF texture size
          if(Npow<13)
          {
	    Npow++;
            texMapSizeY[0] = (int)pow(2.f, Npow);
            texMapSizeX[0] = (int)pow(2.f, Npow);
	    printf("   Texture size : [%4.0f x %4.0f]                                                     \r",texMapSizeY[0],texMapSizeX[0]);fflush(stdout);
	    glutPostRedisplay();
	  }
        break;

      case 'o':        // loop between OBJ models
          actualOBJ++;
          actualOBJ %= loadedBTFObjects;
          glutPostRedisplay();
          printf("   OBJ model: %s (%d)                                                              \r",objectName[actualOBJ],actualOBJ);fflush(stdout);
        break;
      case 'O':        // loop between OBJ models
          actualOBJ--;
	  if (actualOBJ < 0)
	    actualOBJ = loadedBTFObjects-1;
          glutPostRedisplay();
          printf("   OBJ model: %s (%d)                                                               \r",objectName[actualOBJ], actualOBJ);fflush(stdout);
        break;

      case 'n':        // drawing needles
        if(drawNeedles==0)
          { drawNeedles = 1;
            printf("   Object normals rendering enabled.                                                  \r");fflush(stdout);}
        else          {
            drawNeedles = 0;
            printf("   Object normals rendering disabled.                                                 \r");fflush(stdout);}
        glutPostRedisplay();
        break;

     case 'a':         // rotate object left
        OBJrotationFlag = 1;
        objectShift -= PI/phiPerCircle;
        if(objectShift<0)
          objectShift += 2*PI;
        printf("   Object Phi Shift: %3.0f                                                             \r",(180.f/PI)*objectShift);fflush(stdout);
        glutPostRedisplay();
        break;
      case 's':         // rotate object right
        OBJrotationFlag = 2;
        objectShift += PI/phiPerCircle;
        if(objectShift>=2*PI)
          objectShift -= 2*PI;
        printf("   Object Phi Shift: %3.0f                                                             \r",(180.f/PI)*objectShift);fflush(stdout);
        glutPostRedisplay();
        break;
      case 'd':         // rotate object right
        if(animation==0)
          { animation = 1;
            printf("   Object animation enabled.                                                        \r");fflush(stdout);}
        else
          { animation = 0;
            printf("   Object animation disabled.                                                       \r");fflush(stdout);}
        glutPostRedisplay();
        break;
      case 'h':         // printing of help
        printHelp();
        break;

   default:
     break;
   }

  freebmem1(buffer, 0,4*win_w*win_h-1);
  return;
}//--- processKeys ------------------------------------------------

void 
processMouse(int button, int state, int winx, int winy)
// callback function on mouse events
{
	if(button == GLUT_LEFT_BUTTON && state == GLUT_DOWN)
		mb[0] = 1;
	if(button == GLUT_LEFT_BUTTON && state == GLUT_UP)
		mb[0] = 0;
	if(button == GLUT_MIDDLE_BUTTON && state == GLUT_DOWN)
		mb[1] = 1;
	if(button == GLUT_MIDDLE_BUTTON && state == GLUT_UP)
		mb[1] = 0;
	if(button == GLUT_RIGHT_BUTTON && state == GLUT_DOWN)
		mb[2] = 1;
	if(button == GLUT_RIGHT_BUTTON && state == GLUT_UP)
		mb[2] = 0;

    // redraw the scene if any mouse-button was pressed
    if(mb[0]==0 && mb[1]==0 && mb[2]==0)
      {
        glutSwapBuffers();
        glutPostRedisplay();
      }
	if((mb[0]==1)&&(mb[2]==1))
      {  // change view/light mode
        if(lightShiftMode == 1)
          {
            lightShiftMode = 0;
            strcpy(modestring,"camera mode");
            printf("Camera: x: %7.2f y: %7.2f  height: %7.2f\r",cpos[0],cpos[1],cpos[2]);fflush(stdout);
          }
        else
          {
            lightShiftMode = 1;
            strcpy(modestring,"light mode");
            printf("Light : x: %7.2f y: %7.2f  height: %7.2f\r",lpos[0],lpos[1],lpos[2]);fflush(stdout);
          }
      }

	oldwinx = winx;
	oldwiny = winy;
	return;
}//--- processMouse -----------------------------------------------------

void 
motion(int winx, int winy)
// callback function on mouse movement
{
  int i;
  float rotmat[16];
  float rotmatinv[16];

  if(mb[0]){  // rotate
    if(lightShiftMode==0)
      {
        hpr[0] += float(winx - oldwinx) * 0.1f;
        hpr[1] += float(winy - oldwiny) * 0.1f;
      }
    else
      {
        lpos[0] += 2* (winx - oldwinx);
        lpos[1] -= 2* (winy - oldwiny);
      }
  }

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glRotatef(hpr[2], 0.0f, 0.0f, 1.0f);
  glRotatef(hpr[1], 1.0f, 0.0f, 0.0f);
  glRotatef(hpr[0], 0.0f, 1.0f, 0.0f);
  glGetFloatv(GL_MODELVIEW_MATRIX, rotmat);
  invertMatrix(rotmat, rotmatinv);

  if(lightShiftMode==1)
    {
      printf("Light : x: %7.2f y: %7.2f  height: %7.2f\r",lpos[0],lpos[1],lpos[2]);fflush(stdout);
    }
  else
    {
      printf("Camera: x: %7.2f y: %7.2f  height: %7.2f\r",cpos[0],cpos[1],cpos[2]);fflush(stdout);
    }

  if(mb[1]){  // pan
    if(lightShiftMode==0)
      {
        xyz[0] += (float(winx - oldwinx) * rotmatinv[0]
                   - float(winy - oldwiny) * rotmatinv[4]) * 0.8f;//0.01
        xyz[1] += (float(winx - oldwinx) * rotmatinv[1]
                   - float(winy - oldwiny) * rotmatinv[5]) * 0.8f;//0.01
        xyz[2] += (float(winx - oldwinx) * rotmatinv[2]
                   - float(winy - oldwiny) * rotmatinv[6]) * 0.8f;//0.01
        cpos[0] = -xyz[0];
        cpos[1] = -xyz[1];
        cpos[2] = -xyz[2];
      }
    else
      {
        lpos[0] = 0;
        lpos[1] = 0;
        lpos[2] = 0;
      }
  }
  if(mb[2]){  // dolly
    if(lightShiftMode==0)
      {
        xyz[0] += float(winy - oldwiny) * rotmatinv[8] * 0.8f;//0.01
        xyz[1] += float(winy - oldwiny) * rotmatinv[9] * 0.8f;//0.01
        xyz[2] += float(winy - oldwiny) * rotmatinv[10] * 0.8f;//0.01
        cpos[0] = -xyz[0];
        cpos[1] = -xyz[1];
        cpos[2] = -xyz[2];
      }
    else
      lpos[2] -= (winy - oldwiny);
  }
  
  glTranslatef(xyz[0], xyz[1], xyz[2]);
  
  oldwinx = winx;
  oldwiny = winy;
  return;
}//--- motion --------------------------------------------------------------------

void 
setUniformVariables(GLuint program)
// sets shader's uniform/constant variables
{
  GLint loc;

  //=== for BTF shaders ============================================
  // Select texture unit 1 as the active unit and bind the "index" texture. 
  glActiveTexture(GL_TEXTURE1);
  loc = glGetUniformLocationARB(program,"cubemap_ba");
  glUniform1iARB(loc,1);  // Bind indices cubemap to texture unit 1 
  glBindTexture(GL_TEXTURE_CUBE_MAP,cubeMaps[0]);

  glActiveTexture(GL_TEXTURE2);
  loc = glGetUniformLocationARB(program,"cubemap_tv");
  glUniform1iARB(loc,2);  // Bind weights cubemap to texture unit 2 
  glBindTexture(GL_TEXTURE_CUBE_MAP,cubeMaps[1]);
  return;
}//--- setUniformVariables ------------------------------------------------

void 
setShaders() 
// loads, compile and link shaders programs
{
    char *vs,*fs;

    //=== BTF vertex and fragment programs setting up ==================
	v_BTF = glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
	f_BTF = glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB);

	vs = readProgram("../data/shaders/btfbase.vert");
	fs = readProgram("../data/shaders/btfbase.frag");

	const char * ff = fs;
	const char * vv = vs;
	glShaderSourceARB(v_BTF, 1, &vv,NULL);
	glShaderSourceARB(f_BTF, 1, &ff,NULL);

	free(vs);free(fs);

	glCompileShaderARB(v_BTF);
	glCompileShaderARB(f_BTF);

	p_BTF = glCreateProgramObjectARB();
	glAttachObjectARB(p_BTF,f_BTF);
	glAttachObjectARB(p_BTF,v_BTF);

	glLinkProgramARB(p_BTF);

	glUseProgramObjectARB(p_BTF);

    printInfoLog(p_BTF,"BTF shaders");

    setUniformVariables(p_BTF);
    
	//=== color shading vertex and fragment programs setting up ========
	v_color = glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
	f_color = glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB);

	vs = readProgram("../data/shaders/color.vert");
	fs = readProgram("../data/shaders/color.frag");

	ff = fs;
	vv = vs;
	glShaderSourceARB(v_color, 1, &vv,NULL);
	glShaderSourceARB(f_color, 1, &ff,NULL);

	free(vs);free(fs);

	glCompileShaderARB(v_color);
	glCompileShaderARB(f_color);

	p_color = glCreateProgramObjectARB();
	glAttachObjectARB(p_color,f_color);
	glAttachObjectARB(p_color,v_color);

	glLinkProgramARB(p_color);

	glUseProgramObjectARB(p_color);

    printInfoLog(p_color,"color shaders");

	return;
}//--- setShaders --------------------------------------------------------------

void 
setTexParams(GLenum target, GLint clamping, GLint filtering)
// setting texturing params. The same for all textures 
{
  glTexParameteri(target, GL_TEXTURE_WRAP_S, clamping);
  glTexParameteri(target, GL_TEXTURE_WRAP_T, clamping);
  glTexParameteri(target, GL_TEXTURE_MAG_FILTER, filtering);
  glTexParameteri(target, GL_TEXTURE_MIN_FILTER, filtering);
  return;
}//-- setTexParams -------------------------------------------------------------

void
info()
{
  printf("_________________________________________________________________________\n"
         "|                                                                       |\n"
         "|  BTFBASE Shader V1.0 (J.Filip, UTIA AS CR, www.utia.cz, 11/2010)      |\n"
         "|_______________________________________________________________________|\n\n");
  return;
}//--- info --------------------------------------------------------------------

void 
init()
{
  info();
  initRenderer();

  strcpy(modestring,"camera mode");

  printf("##### HARDWARE VERIFICATION ################################################\n");
  //=== OpenGL version ===================================================    
  version = (char *)glGetString(GL_VERSION);
  printf("OpenGL version: %s\n",version);

  //=== test of OpenGL SL support ========================================    
  glewInit();
  if (GLEW_ARB_vertex_shader && GLEW_ARB_fragment_shader)
    printf("Ready for GLSL\n");
  else {
    printf("No GLSL support\nExiting...\n");
    exit(1);
  }

  //=== test of required texturesize and No. of TUs ======================    
  GLint ntex=0;
  glGetIntegerv(GL_MAX_TEXTURE_SIZE,&ntex);
  if(ntex>=2048)
    printf("Max. texture size: %d\n",ntex);
  else
    {
      printf("Insufficient max. texture size! Available: %dpx (required: 2048px)\nExiting...\n",ntex);
      exit(1);
    }
  ntex=0;
  glGetIntegerv(GL_MAX_TEXTURE_UNITS_ARB,&ntex);
  if(ntex>=4)
    printf("Texture units available: %d\n",ntex);
  else
    {
      printf("Insufficient no. of texture units! Available: %d (required: 4)\nExiting...\n",ntex);
      exit(1);
    }
  ntex=0;
  glGetIntegerv(GL_MAX_VERTEX_UNIFORM_COMPONENTS,&ntex);
  printf("GL_MAX_VERTEX_UNIFORM_COMPONENTS: %d\n",ntex);
  glGetIntegerv(GL_MAX_FRAGMENT_UNIFORM_COMPONENTS,&ntex);
  printf("GL_MAX_FRAGMENT_UNIFORM_COMPONENTS: %d\n",ntex);

  
  printf("\n");
  printf("##### CUBEMAPS LOADING #####################################################\n");
  //=== cubemaps definitions and loading =================================
  int cubeMapSize = 256;
  printf("Cubemap resolution: %d\n",cubeMapSize);
  glGenTextures(2, cubeMaps);
  glEnable(GL_TEXTURE_CUBE_MAP);

  int PNGres;
  unsigned char *cubeFaceIndex  = ballocation1(0,cubeMapSize*cubeMapSize*4-1);
  unsigned char *cubeFaceWeight = ballocation1(0,cubeMapSize*cubeMapSize*4-1);
  char fileName[1000];

  glBindTexture(GL_TEXTURE_CUBE_MAP, cubeMaps[0]);
  setTexParams(GL_TEXTURE_CUBE_MAP, GL_REPEAT, GL_NEAREST);
  printf("Loading cubemap - indices+weights (beta,alpha)...\r");fflush(stdout);
  for (int i = 0; i < 6; i++) 
    {
      sprintf(fileName,"../data/cubeMaps/cubeIndices_ba_%d.png",i);
      PNGres = read_PNG_file(fileName, cubeFaceIndex,3, cubeMapSize, cubeMapSize, 0);
      if(PNGres != 1)
        printf("Error: Cubemap face not read (PNGreader: %d)\n",PNGres);

      sprintf(fileName,"../data/cubeMaps/cubeWeights_ba_%d.png",i);
      PNGres = read_PNG_file(fileName, cubeFaceWeight,3, cubeMapSize, cubeMapSize, 0);
      if(PNGres != 1)
        printf("Error: Cubemap face not read (PNGreader: %d)\n",PNGres);

      // copying weights to BA channels of RGBA texture containing indices in RG channels
      for(int irow=0;irow<cubeMapSize;irow++)
        for(int jcol=0;jcol<cubeMapSize;jcol++)
          {
            cubeFaceIndex[4*(irow*cubeMapSize+jcol)+2] = cubeFaceWeight[4*(irow*cubeMapSize+jcol)  ];
            cubeFaceIndex[4*(irow*cubeMapSize+jcol)+3] = cubeFaceWeight[4*(irow*cubeMapSize+jcol)+1];
          }
      glTexImage2D(   cubefaces[i],
                      0,                  // level
                      GL_RGBA8,           // internal format
                      cubeMapSize,        // width
                      cubeMapSize,        // height
                      0,                  // border
                      GL_RGBA,            // format
                      GL_UNSIGNED_BYTE,   // type
                      cubeFaceIndex);     // indices data
    }
  printf("Loading cubemap - indices+weights (beta,alpha)...done\n");fflush(stdout);

  glBindTexture(GL_TEXTURE_CUBE_MAP, cubeMaps[1]);
  setTexParams(GL_TEXTURE_CUBE_MAP, GL_REPEAT, GL_NEAREST);
  printf("Loading cubemap - indices+weights (theta,phi)...\r");fflush(stdout);
  for (int i = 0; i < 6; i++) 
    {
      sprintf(fileName,"../data/cubeMaps/cubeIndices_tp_%d.png",i);
      PNGres = read_PNG_file(fileName, cubeFaceIndex,3, cubeMapSize, cubeMapSize, 0);
      if(PNGres != 1)
        printf("Error: Cubemap face not read (PNGreader: %d)\n",PNGres);

      sprintf(fileName,"../data/cubeMaps/cubeWeights_tp_%d.png",i);
      PNGres = read_PNG_file(fileName, cubeFaceWeight,3, cubeMapSize, cubeMapSize, 0);
      if(PNGres != 1)
        printf("Error: Cubemap face not read (PNGreader: %d)\n",PNGres);

      // copying weights to TP channels of RGBA texture containing indices in RG channels
      for(int irow=0;irow<cubeMapSize;irow++)
        for(int jcol=0;jcol<cubeMapSize;jcol++)
          {
            cubeFaceIndex[4*(irow*cubeMapSize+jcol)+2] = cubeFaceWeight[4*(irow*cubeMapSize+jcol)  ];
            cubeFaceIndex[4*(irow*cubeMapSize+jcol)+3] = cubeFaceWeight[4*(irow*cubeMapSize+jcol)+1];
          }
      glTexImage2D(   cubefaces[i],
                      0,                  // level
                      GL_RGBA,            // internal format
                      cubeMapSize,        // width
                      cubeMapSize,        // height
                      0,                  // border
                      GL_RGBA,            // format
                      GL_UNSIGNED_BYTE,   // type
                      cubeFaceIndex);     // weights data
    }
  printf("Loading cubemap - indices+weights (theta,phi)...done\n");fflush(stdout);

  freebmem1(cubeFaceIndex , 0,cubeMapSize*cubeMapSize*4-1);
  freebmem1(cubeFaceWeight, 0,cubeMapSize*cubeMapSize*4-1);

  printf("\n");
  printf("##### GPU SHADERS LOADING ##################################################\n");

  //=== integer textures allocation =============================================
  glGenTextures(4, indexTex);

  //=== float textures allocation ===============================================
  glGenTextures(2, scaleTex);

  //=== set rendering options ===================================================
  glClearColor(0.220, 0.259, 0.251, 1.0);
  glPolygonMode(GL_FRONT,GL_FILL);
  glShadeModel(GL_SMOOTH); //GL_SMOOTH,GL_FLAT
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_CULL_FACE);
  glEnable(GL_LIGHTING);

  //=== setting shaders ============================================================
  setShaders();
  return;
}//--- init ------------------------------------------------------

int 
addMaterial()
{
  NoOfLoadedMaterials++;
  int materialNumber = NoOfLoadedMaterials;

  int testPar;

  //=== TI16_1 proxy testing, definitions and loading =================================
  glTexImage2D(GL_PROXY_TEXTURE_2D, 0, GL_RGBA16, texSize,texSize, 0, GL_RGBA, GL_UNSIGNED_SHORT, NULL);
  glGetTexLevelParameteriv(GL_PROXY_TEXTURE_2D, 0, GL_TEXTURE_WIDTH, &testPar);
  if(testPar==0)
    { printf("Texture can't be accommodated in memory!!!\nExiting...\n");
      exit(0); }

  glBindTexture(GL_TEXTURE_2D, indexTex[0]);
  setTexParams(GL_TEXTURE_2D, GL_REPEAT, GL_NEAREST);
  glTexImage2D(   GL_TEXTURE_2D,
                  0,                  // level
                  GL_RGBA16,          // internal format
                  texSize,            // width
                  texSize,            // height
                  0,                  // border
                  GL_RGBA,            // format
                  GL_UNSIGNED_SHORT,  // type
                  BTF->TI16_1);       // pixel data
 
  //=== TI16_2 proxy testing, definitions and loading =================================
  glTexImage2D(GL_PROXY_TEXTURE_2D, 0, GL_RGBA16, texSize,texSize, 0, GL_RGBA, GL_UNSIGNED_SHORT, NULL);
  glGetTexLevelParameteriv(GL_PROXY_TEXTURE_2D, 0, GL_TEXTURE_WIDTH, &testPar);
  if(testPar==0)
    { printf("Texture can't be accommodated in memory!!!\nExiting...\n");
      exit(0); }

  glBindTexture(GL_TEXTURE_2D, indexTex[1]);
  setTexParams(GL_TEXTURE_2D, GL_REPEAT, GL_NEAREST);
  glTexImage2D(   GL_TEXTURE_2D,
                  0,                  // level
                  GL_RGBA16,          // internal format
                  texSize,            // width
                  texSize,            // height
                  0,                  // border
                  GL_RGBA,            // format
                  GL_UNSIGNED_SHORT,  // type
                  BTF->TI16_2);       // pixel data

  //=== TI16_3 proxy testing, definitions and loading =================================
  glTexImage2D(GL_PROXY_TEXTURE_2D, 0, GL_RGBA16, texSize,texSize, 0, GL_RGBA, GL_UNSIGNED_SHORT, NULL);
  glGetTexLevelParameteriv(GL_PROXY_TEXTURE_2D, 0, GL_TEXTURE_WIDTH, &testPar);
  if(testPar==0)
    { printf("Texture can't be accommodated in memory!!!\nExiting...\n");
      exit(0); }

  glBindTexture(GL_TEXTURE_2D, indexTex[2]);
  setTexParams(GL_TEXTURE_2D, GL_REPEAT, GL_NEAREST);
  glTexImage2D(   GL_TEXTURE_2D,
                  0,                  // level
                  GL_RGBA16,          // internal format
                  texSize,            // width
                  texSize,            // height
                  0,                  // border
                  GL_RGBA,            // format
                  GL_UNSIGNED_SHORT,  // type
                  BTF->TI16_3);       // pixel data
 
  //=== TI16_4 proxy testing, definitions and loading =================================
  glTexImage2D(GL_PROXY_TEXTURE_2D, 0, GL_RGBA16, texSize,texSize, 0, GL_RGBA, GL_UNSIGNED_SHORT, NULL);
  glGetTexLevelParameteriv(GL_PROXY_TEXTURE_2D, 0, GL_TEXTURE_WIDTH, &testPar);
  if(testPar==0)
    { printf("Texture can't be accommodated in memory!!!\nExiting...\n");
      exit(0); }

  glBindTexture(GL_TEXTURE_2D, indexTex[3]);
  setTexParams(GL_TEXTURE_2D, GL_REPEAT, GL_NEAREST);
  glTexImage2D(   GL_TEXTURE_2D,
                  0,                  // level
                  GL_RGBA16,          // internal format
                  texSize,            // width
                  texSize,            // height
                  0,                  // border
                  GL_RGBA,            // format
                  GL_UNSIGNED_SHORT,  // type
                  BTF->TI16_4);       // pixel data

  //=== TF32_scales proxy testing, definitions and loading ============================
  glTexImage2D(GL_PROXY_TEXTURE_2D, 0, GL_RGBA32F_ARB, texSize,texSize, 0, GL_RGBA, GL_FLOAT, NULL);
  glGetTexLevelParameteriv(GL_PROXY_TEXTURE_2D, 0, GL_TEXTURE_WIDTH, &testPar);
  if(testPar==0)
    { printf("Texture can't be accommodated in memory!!!\nExiting...\n");
      exit(0); }

  glBindTexture(GL_TEXTURE_2D, scaleTex[0]);
  setTexParams(GL_TEXTURE_2D, GL_REPEAT, GL_NEAREST);
  glTexImage2D(   GL_TEXTURE_2D,
                  0,                  // level
                  GL_RGBA16F_ARB,     // internal format
                  texSize,            // width
                  texSize,            // height
                  0,                  // border
                  GL_RGBA,            // format
                  GL_FLOAT,           // type
                  BTF->TF32_scales);  // pixel data
 
  //=== TF32_data proxy testing, definitions and loading =============================
  glTexImage2D(GL_PROXY_TEXTURE_2D, 0, GL_RGBA32F_ARB, texSize,texSize, 0, GL_RGBA, GL_FLOAT, NULL);
  glGetTexLevelParameteriv(GL_PROXY_TEXTURE_2D, 0, GL_TEXTURE_WIDTH, &testPar);
  if(testPar==0)
    { printf("Texture can't be accommodated in memory!!!\nExiting...\n");
      exit(0); }

  glBindTexture(GL_TEXTURE_2D, scaleTex[1]);
  setTexParams(GL_TEXTURE_2D, GL_REPEAT, GL_NEAREST);
  glTexImage2D(   GL_TEXTURE_2D,
                  0,                  // level
                  GL_RGBA16F_ARB,     // internal format
                  texSize,            // width
                  texSize,            // height
                  0,                  // border
                  GL_RGBA,            // format
                  GL_FLOAT,           // type
                  BTF->TF32_data);    // pixel data

  BTF->deallocTextures(); // dealloc material parametric textures 
  return materialNumber;
}//--- addMaterial -----------------------------------------------

void
printUsage()
{
  printf("Usage:                                          \n"
         "        lightenvGPU [materialPath]              \n"
         "        lightenvGPU h         - for help        \n\n");
  return;
}//--- printUsage ------------------------------------------------


//#####################################################################################
//### MAIN ############################################################################
//#####################################################################################

int 
main(int argc, char **argv) 
{
  if (argc < 2) {
    printUsage();
    exit(0);
  }
  if (strcmp(argv[1],"h") == 0) {
    printHelp();
    exit(0);
  }

  //=== allocations ==============================================
  auxString = callocation1(0,300);
  texMapSizeX = allocation1(0,NoOfObjects-1);
  texMapSizeY = allocation1(0,NoOfObjects-1);
  objectMaterialID = iallocation1(0,NoOfObjects-1);

  // initialize codebooks sizes
  BTF = new BTFtools(texSize, SlicesPerPhi, SlicesPerTheta, SlicesPerHemi, LengthOfSlice);  
  //=== object data allocation ===================================
  for(int i=0;i<NoOfObjects;i++)
    DATA[i] = new TObjectData(verticesPerObject);  

  //=== glut functions assignment  ===============================
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
  glutInitWindowPosition(100,100);
  glutInitWindowSize(win_w,win_h);
  glutCreateWindow("BTFBASE Renderer");
  
  glutDisplayFunc(renderScene);
  glutIdleFunc(renderScene);
  glutReshapeFunc(changeSize);
  glutKeyboardFunc(processKeys);
  glutMouseFunc(processMouse);
  glutMotionFunc(motion);

  //=== OpenGL+GLSL initialisation ===============================
  init(); 
  //==== GEOMETRY ==============================================================================
  printf("\n");
  printf("##### GEOMETRY LOADING #####################################################\n");
   
  //###  one centered model rendering ######################
  // sphere ========================
  DATA[loadedBTFObjects]->loadOBJfile("../data/obj/koule.obj", 1, 1.f, 0.f,0.f,0.f);
  DATA[loadedBTFObjects]->tangentsPerVertex();
      
  strcpy(objectName[loadedBTFObjects],"sphere");
  texMapSizeY[loadedBTFObjects] = (int)pow(2.f, Npow); // here can be arbitrary number (no 2^n restriction)
  texMapSizeX[loadedBTFObjects] = (int)pow(2.f, Npow);
  loadedBTFObjects++;

  // bunny ========================
  DATA[loadedBTFObjects]->loadOBJfile("../data/obj/bunny.obj", 1, 13.f, 0.f,0.f,0.f);
  DATA[loadedBTFObjects]->tangentsPerVertex();

  strcpy(objectName[loadedBTFObjects],"bunny");
  texMapSizeY[loadedBTFObjects] = (int)pow(2.f, Npow); // here can be arbitrary number (no 2^n restriction)
  texMapSizeX[loadedBTFObjects] = (int)pow(2.f, Npow);
  loadedBTFObjects++;

  //=== loading BTF materials ====================================
  // load BTFBASE params. into textures
  printf("\n");
  printf("##### BTFBASE PARAMETERS ###################################################\n");
  BTF->setMaterial(argv[1],BTFwidth,BTFheight, hdrMultiplier);
  int matID = addMaterial();
  objectMaterialID[0] = matID;
  objectMaterialID[1] = matID;

  //=== RENDERING ================================================
  printf("\n");
  printf("##### GPU RENDERING #####################################################>>>\n");
  glutMainLoop();

  //=== deallocations ============================================
  freecmem1(auxString, 0,300);
  freemem1(texMapSizeX, 0,NoOfObjects-1);
  freemem1(texMapSizeY, 0,NoOfObjects-1);
  freeimem1(objectMaterialID, 0,NoOfObjects-1);

  return 0;
}//--- main ---------------------------------------------------------

